# coding: utf-8
"""
Global Named Events.
"""
from __future__ import annotations
from typing import NamedTuple


class GblNamedEvent(NamedTuple):
    """
    Global Named Events
    """

    PRINTING = "global_printing"
    """Global Printing Event."""
